#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import curses
from contextlib import contextmanager
from mock import patch
import test_common

from patsi.document import color, palette
from patsi import widgets
from patsi.widgets import geometry as geo
from patsi.widgets import _util as util


class TestPoint(test_common.TestCase):
    def assert_point(self, point, x, y):
        self.assertEqual(point.x, x)
        self.assertEqual(point.y, y)

    def test_ctor(self):
        self.assert_point(geo.Point(), 0, 0)
        self.assert_point(geo.Point(1, 2), 1, 2)
        self.assert_point(geo.Point(1, 2, reversed=True), 2, 1)
        self.assert_point(geo.Point((1, 2)), 1, 2)
        self.assert_point(geo.Point(x=1, y=2), 1, 2)

    def test_add(self):
        self.assert_point(geo.Point(1, 2) + geo.Point(30, 40), 31, 42)

        point = geo.Point(1, 2)
        point += geo.Point(30, 40)
        self.assert_point(point, 31, 42)

        self.assert_point(geo.Point(34, 43) - geo.Point(3, 1), 31, 42)

        point = geo.Point(34, 43)
        point -= geo.Point(3, 1)
        self.assert_point(point, 31, 42)

    def test_cmp(self):
        self.assertTrue(geo.Point(1, 2) == geo.Point(1, 2))
        self.assertFalse(geo.Point(1, 2) == geo.Point(1, 3))
        self.assertFalse(geo.Point(3, 2) == geo.Point(1, 2))

        self.assertFalse(geo.Point(1, 2) != geo.Point(1, 2))
        self.assertTrue(geo.Point(1, 2) != geo.Point(1, 3))
        self.assertTrue(geo.Point(3, 2) != geo.Point(1, 2))

    def test_interpolate(self):
        self.assert_point(geo.Point(4, 5).interpolate(geo.Point(2, 3)), 3, 4)
        self.assert_point(geo.Point(4, 5).interpolate(geo.Point(2, 3), 0), 4, 5)
        self.assert_point(geo.Point(4, 5).interpolate(geo.Point(2, 3), 1), 2, 3)

    def test_elements(self):
        point = geo.Point(4, 5)
        self.assertEqual(point[0], point.x)
        self.assertEqual(point["x"], point.x)
        self.assertEqual(point[1], point.y)
        self.assertEqual(point["y"], point.y)
        self.assertRaises(KeyError, lambda: point[2])
        self.assertRaises(KeyError, lambda: point["z"])

        x, y = point
        self.assert_point(point, x, y)
        self.assertEqual(tuple(point), (x, y))

    def test_repr(self):
        point = geo.Point(4, 5)
        self.assertEqual(str(point), str(tuple(point)))


class TestRect(test_common.TestCase):
    def assert_rect(self, rect, x1, y1, x2, y2):
        self.assertEqual(rect.top_left, geo.Point(x1, y1))
        self.assertEqual(rect.bottom_right, geo.Point(x2, y2))

    def test_ctor(self):
        tl = geo.Point(1, 2)
        br = geo.Point(11, 22)
        sz = geo.Point(10, 20)
        self.assert_rect(geo.Rect(pos=tl, size=sz), 1, 2, 11, 22)
        self.assert_rect(geo.Rect(top_left=tl, bottom_right=br), 1, 2, 11, 22)
        self.assert_rect(
            geo.Rect(x=tl.x, y=tl.y, width=sz.x, height=sz.y),
            1, 2, 11, 22
        )
        self.assert_rect(
            geo.Rect(x1=tl.x, y1=tl.y, x2=br.x, y2=br.y),
            1, 2, 11, 22
        )
        self.assert_rect(
            geo.Rect(top=tl.x, left=tl.y, bottom=br.x, right=br.y),
            1, 2, 11, 22
        )
        self.assert_rect(geo.Rect(), 0, 0, 0, 0)

    def test_contains(self):
        rect = geo.Rect(x1=10, y1=10, x2=20, y2=20)
        pin = geo.Point(12, 13)
        pout = geo.Point(14, 2)
        pbound = geo.Point(14, 10)

        self.assertTrue(rect.contains(pin))
        self.assertFalse(rect.contains(pout))
        self.assertTrue(rect.contains(pbound))

        self.assertTrue(rect.contains(pin, False))
        self.assertFalse(rect.contains(pout, False))
        self.assertFalse(rect.contains(pbound, False))

    def test_property_getters(self):
        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        self.assertEqual(rect.top_left, geo.Point(1, 2))
        self.assertEqual(rect.bottom_right, geo.Point(11, 32))
        self.assertEqual(rect.top_right, geo.Point(11, 2))
        self.assertEqual(rect.bottom_left, geo.Point(1, 32))
        self.assertEqual(rect.center, geo.Point(6, 17))

        self.assertEqual(rect.top, 2)
        self.assertEqual(rect.left, 1)
        self.assertEqual(rect.bottom, 32)
        self.assertEqual(rect.right, 11)

        self.assertEqual(rect.x1, 1)
        self.assertEqual(rect.y1, 2)
        self.assertEqual(rect.x2, 11)
        self.assertEqual(rect.y2, 32)

        self.assertEqual(rect.x, 1)
        self.assertEqual(rect.y, 2)
        self.assertEqual(rect.width, 10)
        self.assertEqual(rect.height, 30)

        self.assertEqual(rect.pos, geo.Point(1, 2))
        self.assertEqual(rect.size, geo.Point(10, 30))

    def test_property_setters(self):
        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.top_left = geo.Point(10, 20)
        self.assert_rect(rect, 10, 20, 11, 32)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.bottom_right = geo.Point(30, 40)
        self.assert_rect(rect, 1, 2, 30, 40)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.top_right = geo.Point(10, 20)
        self.assert_rect(rect, 1, 20, 10, 32)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.bottom_left = geo.Point(40, 20)
        self.assert_rect(rect, 40, 2, 11, 20)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.left = 4
        rect.top = 5
        rect.right = 6
        rect.bottom = 7
        self.assert_rect(rect, 4, 5, 6, 7)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.x = 101
        rect.y = 102
        self.assert_rect(rect, 101, 102, 111, 132)


        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.x1 = 101
        rect.y1 = 102
        self.assert_rect(rect, 101, 102, 11, 32)
        rect.x2 = 201
        rect.y2 = 202
        self.assert_rect(rect, 101, 102, 201, 202)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.pos = geo.Point(101, 102)
        self.assert_rect(rect, 101, 102, 111, 132)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.width = 100
        rect.height = 200
        self.assert_rect(rect, 1, 2, 101, 202)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.size = geo.Point(100, 200)
        self.assert_rect(rect, 1, 2, 101, 202)

        rect = geo.Rect(x1=1, y1=2, x2=11, y2=32)
        rect.center = geo.Point(0, 0)
        self.assert_rect(rect, -5, -15, 5, 15)

    def test_comparison(self):
        self.assertTrue(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) ==
            geo.Rect(x1=10, y1=20, x2=30, y2=40)
        )
        self.assertFalse(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) ==
            geo.Rect(x1=15, y1=20, x2=30, y2=40)
        )
        self.assertFalse(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) ==
            geo.Rect(x1=10, y1=25, x2=30, y2=40)
        )
        self.assertFalse(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) ==
            geo.Rect(x1=10, y1=20, x2=35, y2=40)
        )
        self.assertFalse(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) ==
            geo.Rect(x1=10, y1=20, x2=30, y2=45)
        )

        self.assertFalse(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) !=
            geo.Rect(x1=10, y1=20, x2=30, y2=40)
        )
        self.assertTrue(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) !=
            geo.Rect(x1=15, y1=20, x2=30, y2=40)
        )
        self.assertTrue(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) !=
            geo.Rect(x1=10, y1=25, x2=30, y2=40)
        )
        self.assertTrue(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) !=
            geo.Rect(x1=10, y1=20, x2=35, y2=40)
        )
        self.assertTrue(
            geo.Rect(x1=10, y1=20, x2=30, y2=40) !=
            geo.Rect(x1=10, y1=20, x2=30, y2=45)
        )

    def test_copy(self):
        r1 = geo.Rect(x1=10, y1=20, x2=30, y2=40)
        r2 = r1.copy()
        self.assertIsNot(r1, r2)
        r2.x1 = 100
        r2.y1 = 200
        r2.x2 = 300
        r2.y2 = 400
        self.assert_rect(r1, 10, 20, 30, 40)
        self.assert_rect(r2, 100, 200, 300, 400)

    def test_repr(self):
        rect = geo.Rect(x=10, y=20, width=30, height=40)
        self.assertEqual(str(rect), "Rect(30x40+10+20)")


class TestUtil(test_common.TestCase):
    def test_next_object(self):
        arr = [0, 1, 2, 3, 4]
        curr = 2
        self.assertEqual(4, util.next_object(arr, curr, -3))
        self.assertEqual(0, util.next_object(arr, curr, -2))
        self.assertEqual(1, util.next_object(arr, curr, -1))
        self.assertEqual(2, util.next_object(arr, curr, 0))
        self.assertEqual(3, util.next_object(arr, curr, 1))
        self.assertEqual(4, util.next_object(arr, curr, 2))
        self.assertEqual(0, util.next_object(arr, curr, 3))

        curr = None
        self.assertEqual(2, util.next_object(arr, curr, -3))
        self.assertEqual(3, util.next_object(arr, curr, -2))
        self.assertEqual(4, util.next_object(arr, curr, -1))
        self.assertEqual(0, util.next_object(arr, curr, 0))
        self.assertEqual(0, util.next_object(arr, curr, 1))
        self.assertEqual(1, util.next_object(arr, curr, 2))
        self.assertEqual(2, util.next_object(arr, curr, 3))

        self.assertEqual(curr, util.next_object([], curr, 3))

    @patch("curses.color_pair", side_effect=lambda x: 256 << (x-1) if x else 0)
    def test_color_to_curses(self, color_pair):
        color_conv = widgets.curses_helper.color_to_curses

        self.assertEqual(
            curses.A_NORMAL | curses.color_pair(0),
            color_conv(None)
        )

        self.assertEqual(
            curses.A_NORMAL | curses.color_pair(2),
            color_conv(color.IndexedColor(1, palette.colors16))
        )

        self.assertEqual(
            curses.A_BOLD | curses.color_pair(2),
            color_conv(color.IndexedColor(8|1, palette.colors16))
        )

    def test_curses_context(self):
        self.assertRaises(curses.error, lambda: curses.color_pair(1))
        with widgets.curses_helper.curses_context():
            self.assertFalse(curses.isendwin())
        self.assertTrue(curses.isendwin())


class MockCursesWindow(object):
    def __init__(self, nlines, ncols, begin_y=0, begin_x=0):
        self.rect = geo.Rect(x=begin_x, y=begin_y, width=ncols, height=nlines)
        self.strings = {}
        self.input_queue = []
        self.refreshed = False
        self.resized = False

    def subwin(self, nlines, ncols, begin_y, begin_x):
        return MockCursesWindow(nlines, ncols, begin_y, begin_x)

    def addstr(self, y, x, text, flags=0):
        self.strings[(x, y)] = (text, flags)

    def border(self, *args):
        pass

    def getbegyx(self):
        return (self.rect.top, self.rect.left)

    def getmaxyx(self):
        return (self.rect.bottom, self.rect.right)

    def getch(self):
        return self.input_queue.pop(0)

    def mvwin(self, new_y, new_x):
        self.rect.top_left = geo.Point(new_x, new_y)

    def resize(self, nlines, ncols):
        self.rect.size = geo.Point(ncols, nlines)
        self.resized = True

    def clear(self):
        self.strings.clear()
        self.refreshed = False

    def refresh(self):
        self.refreshed = True


class TestEvent(test_common.TestCase):
    def test_ctor(self):
        ev = widgets.Event(123, foo="bar")
        self.assertEqual(ev.type, 123)
        self.assertTrue(hasattr(ev, "foo"))
        self.assertEqual(ev.foo, "bar")
        self.assertTrue(ev.propagate)

    def test_accept(self):
        ev = widgets.Event(123)
        ev.accept()
        self.assertFalse(ev.propagate)

    def test_static(self):
        ev = widgets.Event.text_event("x")
        self.assertEqual(ev.type, widgets.Event.TYPE_TEXT)
        self.assertEqual(ev.char, "x")

        ev = widgets.Event.key_event(123)
        self.assertEqual(ev.type, widgets.Event.TYPE_KEY)
        self.assertEqual(ev.key, 123)

        ev = widgets.Event.resize_event(geo.Rect())
        self.assertEqual(ev.type, widgets.Event.TYPE_RESIZE)
        self.assertEqual(ev.bounds, geo.Rect())

        ev = widgets.Event.mouse_event(geo.Point(), 0)
        self.assertEqual(ev.type, widgets.Event.TYPE_MOUSE)
        self.assertEqual(ev.pos, geo.Point())
        self.assertEqual(ev.buttons, 0)

        ev = widgets.Event.focus_event(True)
        self.assertEqual(ev.type, widgets.Event.TYPE_FOCUS)
        self.assertEqual(ev.focus, True)


class TestWidget(test_common.TestCase):
    @patch("curses.newwin", new=lambda *args: MockCursesWindow(*args))
    def windows(self):
        parent = widgets.Widget(None)
        child = widgets.Widget(parent)
        return parent, child

    @contextmanager
    def no_event(self, widget, type=None):
        old = widget.event

        def noevent(event):
            if type is None or type == event.type:
                raise Exception("Event should not have been called")
            return old(event)

        widget.event = noevent
        try:
            yield
        finally:
            widget.event = old

    @contextmanager
    def must_event(self, widget, type, count=1):
        old = widget.event
        events = []

        def noevent(event):
            if type == event.type:
                events.append(event)
            return old(event)

        widget.event = noevent
        try:
            yield
        finally:
            widget.event = old
            self.assertEqual(len(events), count)

    @patch("curses.newwin", new=lambda *args: MockCursesWindow(*args))
    def test_ctor(self):
        window = MockCursesWindow(640, 480, 1, 2)
        wid = widgets.Widget(None, window)
        self.assertIsNone(wid.parent)
        self.assertEqual(wid.children, [])
        self.assertIsNone(wid.focus_child)
        self.assertTrue(wid.active)
        self.assertIs(wid.window, window)

        wid0 = widgets.Widget(None)
        self.assertIsNone(wid0.parent)
        self.assertEqual(wid0.children, [])
        self.assertIsNone(wid0.focus_child)
        self.assertTrue(wid0.active)
        self.assertIsInstance(wid0.window, MockCursesWindow)

        wid1 = widgets.Widget(wid)
        self.assertIs(wid1.parent, wid)
        self.assertIn(wid1, wid.children)
        self.assertEqual(wid1.children, [])
        self.assertIsNone(wid1.focus_child)
        self.assertTrue(wid1.active)
        self.assertIsInstance(wid1.window, MockCursesWindow)
        self.assertIsNot(wid1.window, wid.window)
        self.assertEqual(wid1.window.rect, wid.window.rect)

        window2 = MockCursesWindow(1, 2, 3, 4)
        wid2 = widgets.Widget(wid, window2)
        self.assertIs(wid2.parent, wid)
        self.assertIn(wid2, wid.children)
        self.assertEqual(wid2.children, [])
        self.assertIsNone(wid2.focus_child)
        self.assertTrue(wid2.active)
        self.assertIsInstance(wid2.window, MockCursesWindow)
        self.assertIs(wid2.window, window2)
        self.assertEqual(wid2.window.rect, wid.window.rect)

    def test_has_focus(self):
        parent, child = self.windows()

        self.assertTrue(parent.has_focus())
        parent.active = False
        self.assertFalse(parent.has_focus())
        parent.active = True

        self.assertFalse(child.has_focus())
        parent.focus(child)
        self.assertTrue(child.has_focus())
        child.active = False
        self.assertFalse(child.has_focus())
        child.active = True
        parent.focus(None)
        self.assertFalse(child.has_focus())

    def test_focus(self):
        parent, child = self.windows()

        parent.active = False
        parent.focus(child)
        self.assertFalse(child.has_focus())
        parent.active = True

        child.focus()
        self.assertTrue(child.has_focus())
        parent.focus(None)

        child.focus()
        with self.no_event(child):
            child.focus()
        parent.focus(None)

        sibling = widgets.Widget(parent)
        child.focus()
        self.assertTrue(child.has_focus())
        self.assertFalse(sibling.has_focus())
        with self.must_event(child, widgets.Event.TYPE_FOCUS):
            sibling.focus()
            self.assertTrue(sibling.has_focus())
            self.assertFalse(child.has_focus())
        parent.focus(None)

        parent.focus()

    def test_deactivate(self):
        parent, child = self.windows()
        self.assertTrue(parent.active)
        parent.deactivate()
        self.assertFalse(parent.active)

    @patch("curses.getmouse", new=lambda: (0, 1, 2, 0, 123))
    def test_get_input(self):
        parent, child = self.windows()
        parent.window.input_queue = [
            curses.KEY_RESIZE,
            curses.KEY_MOUSE,
            ord(" "),
            curses.KEY_HOME,
            curses.KEY_MOUSE,
        ]

        def event_replace(event):
            parent._last_event = event
        parent.event = event_replace

        parent.get_input()
        self.assertEqual(parent._last_event.type, widgets.Event.TYPE_RESIZE)
        self.assertEqual(parent._last_event.bounds, parent.window_bounds())

        parent.get_input()
        self.assertEqual(parent._last_event.type, widgets.Event.TYPE_MOUSE)
        self.assertEqual(parent._last_event.pos, geo.Point(1, 2))
        self.assertEqual(parent._last_event.buttons, 123)

        parent.get_input()
        self.assertEqual(parent._last_event.type, widgets.Event.TYPE_TEXT)
        self.assertEqual(parent._last_event.char, " ")

        parent.get_input()
        self.assertEqual(parent._last_event.type, widgets.Event.TYPE_KEY)
        self.assertEqual(parent._last_event.key, curses.KEY_HOME)

        def getmouse_error():
            raise curses.error()
        with patch("curses.getmouse", new=getmouse_error):
            parent.get_input()
            self.assertIsNotNone(parent._last_event)
            self.assertEqual(parent._last_event.type, widgets.Event.TYPE_KEY)

    def test_default_event_handlers(self):
        parent, child = self.windows()

        pre = vars(parent)
        parent.text_event(widgets.Event.text_event("x"))
        self.assertEqual(pre, vars(parent))

        pre = vars(parent)
        parent.key_event(widgets.Event.key_event(123))
        self.assertEqual(pre, vars(parent))

        self.assertFalse(parent.window.resized)
        bounds = geo.Rect(x1=1, y1=2, x2=11, y2=22)
        parent.resize_event(widgets.Event.resize_event(bounds))
        self.assertEqual(parent.window_bounds(), bounds)
        self.assertTrue(parent.window.resized)
        parent.window.resized = False
        parent.resize_event(widgets.Event.resize_event(bounds))
        self.assertFalse(parent.window.resized)

        self.assertFalse(child.has_focus())
        child.mouse_event(widgets.Event.mouse_event(geo.Point(), 0))
        self.assertTrue(child.has_focus())

        self.assertFalse(parent.window.refreshed)
        parent.focus_event(widgets.Event.focus_event(True))
        self.assertTrue(parent.window.refreshed)
        parent.window.refreshed = False
        parent.focus_event(widgets.Event.focus_event(False))
        self.assertFalse(parent.window.refreshed)

    def test_event_change_focus_child(self):
        parent, child = self.windows()
        self.assertTrue(parent.event_change_focus_child(child))
        self.assertIs(parent.focus_child, child)

    def test_event_change_focus_child(self):
        parent, child = self.windows()

        self.assertFalse(parent.window.refreshed)
        self.assertFalse(child.window.refreshed)
        parent.refresh()
        self.assertTrue(parent.window.refreshed)
        self.assertTrue(child.window.refreshed)

        parent.window.refreshed = False
        child.window.refreshed = False
        child.active = False
        parent.refresh()
        self.assertTrue(parent.window.refreshed)
        self.assertFalse(child.window.refreshed)

    @patch("curses.newwin", new=lambda *args: MockCursesWindow(*args))
    def test_loop(self):
        widget = widgets.Widget(None)
        widget._renders = 0
        def render():
            widget._renders += 1
            if widget._renders > 4:
                raise Exception("Too many iterations")
        widget.render = render
        widget.window.input_queue = [
            curses.KEY_MOUSE,
            curses.KEY_MOUSE,
            curses.KEY_MOUSE,
            ord("q")
        ]
        widget.text_event = lambda *args: widget.deactivate()
        widget.loop()
        self.assertTrue(widget.window.refreshed)
        self.assertEqual(widget._renders, 4)


class TestWidgetEvent(test_common.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestWidgetEvent, self).__init__(*args, **kwargs)

        with patch("curses.newwin", new=lambda *args: MockCursesWindow(*args)):
            self.parent = widgets.Widget(None)
            self.child = widgets.Widget(self.parent)
            self.sibling = widgets.Widget(self.parent)

        self.resize_event= CaptureEvent(self.parent, "resize_event")
        self.text_event  = CaptureEvent(self.parent, "text_event")
        self.key_event   = CaptureEvent(self.parent, "key_event")
        self.mouse_event = CaptureEvent(self.parent, "mouse_event")
        self.focus_event = CaptureEvent(self.parent, "focus_event")

        self.child_resize_event= CaptureEvent(self.child, "resize_event")
        self.child_text_event  = CaptureEvent(self.child, "text_event")
        self.child_key_event   = CaptureEvent(self.child, "key_event")
        self.child_mouse_event = CaptureEvent(self.child, "mouse_event")
        self.child_focus_event = CaptureEvent(self.child, "focus_event")

        self.sibling_resize_event= CaptureEvent(self.sibling, "resize_event")
        self.sibling_text_event  = CaptureEvent(self.sibling, "text_event")
        self.sibling_key_event   = CaptureEvent(self.sibling, "key_event")
        self.sibling_mouse_event = CaptureEvent(self.sibling, "mouse_event")
        self.sibling_focus_event = CaptureEvent(self.sibling, "focus_event")

    def setUp(self):
        self.parent.active = self.child.active = True
        self.parent.focus(self.child)

        self.resize_event.propagate = True
        self.text_event.propagate = True
        self.key_event.propagate = True
        self.mouse_event.propagate = True
        self.focus_event.propagate = True

        self.child_resize_event.propagate = True
        self.child_text_event.propagate = True
        self.child_key_event.propagate = True
        self.child_mouse_event.propagate = True
        self.child_focus_event.propagate = True

        self.sibling_resize_event.propagate = True
        self.sibling_text_event.propagate = True
        self.sibling_key_event.propagate = True
        self.sibling_mouse_event.propagate = True
        self.sibling_focus_event.propagate = True

        self.parent._captured_events = []
        self.child._captured_events = []
        self.sibling._captured_events = []

    def no_focus(self):
        self.parent.focus(None)
        self.child._captured_events = []
        self.sibling._captured_events = []

    def test_not_active(self):
        self.parent.active = False
        self.parent.event(widgets.Event.resize_event(geo.Rect()))
        self.parent.event(widgets.Event.text_event("x"))
        self.parent.event(widgets.Event.key_event(123))
        self.parent.event(widgets.Event.mouse_event(geo.Point(), 123))
        self.parent.event(widgets.Event.focus_event(False))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(self.parent._captured_events[0][0], "resize_event")
        self.parent.active = True

    def test_unknown_event(self):
        self.parent.event(widgets.Event(-1))
        self.assertEqual(self.parent._captured_events, [])

    def test_focus(self):
        self.parent.event(widgets.Event.focus_event(False))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "focus_event")
        self.assertEqual(self.parent._captured_events[0][1].focus, False)

    def test_key_nofocus(self):
        self.no_focus()
        self.parent.event(widgets.Event.key_event(123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "key_event")
        self.assertEqual(self.parent._captured_events[0][1].key, 123)

    def test_key_propagate(self):
        self.parent.event(widgets.Event.key_event(123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 1)
        self.assertEqual(self.parent._captured_events[0][0], "key_event")
        self.assertEqual(self.parent._captured_events[0][1].key, 123)
        self.assertEqual(self.child._captured_events[0][0], "key_event")
        self.assertEqual(self.child._captured_events[0][1].key, 123)

    def test_key_nopropagate(self):
        self.key_event.propagate = False
        self.parent.event(widgets.Event.key_event(123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "key_event")
        self.assertEqual(self.parent._captured_events[0][1].key, 123)

    def test_text_nofocus(self):
        self.no_focus()
        self.parent.event(widgets.Event.text_event("1"))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "text_event")
        self.assertEqual(self.parent._captured_events[0][1].char, "1")

    def test_text_propagate(self):
        self.parent.event(widgets.Event.text_event("1"))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 1)
        self.assertEqual(self.parent._captured_events[0][0], "text_event")
        self.assertEqual(self.parent._captured_events[0][1].char, "1")
        self.assertEqual(self.child._captured_events[0][0], "text_event")
        self.assertEqual(self.child._captured_events[0][1].char, "1")

    def test_text_nopropagate(self):
        self.text_event.propagate = False
        self.parent.event(widgets.Event.text_event("1"))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "text_event")
        self.assertEqual(self.parent._captured_events[0][1].char, "1")

    def test_mouse_nopropagate(self):
        self.no_focus()
        self.mouse_event.propagate = False
        self.parent.event(widgets.Event.mouse_event(geo.Point(1, 2), 123))

        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)

        self.assertEqual(self.parent._captured_events[0][0], "mouse_event")
        self.assertEqual(self.parent._captured_events[0][1].pos, geo.Point(1, 2))
        self.assertEqual(self.parent._captured_events[0][1].buttons, 123)

        self.assertFalse(self.child.has_focus())

    def test_mouse_nofocus(self):
        self.no_focus()
        self.child_mouse_event.propagate = False
        self.parent.event(widgets.Event.mouse_event(geo.Point(1, 2), 123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 1)
        self.assertEqual(len(self.sibling._captured_events), 0)

        self.assertEqual(self.child._captured_events[0][0], "mouse_event")
        self.assertEqual(self.child._captured_events[0][1].pos, geo.Point(1, 2))
        self.assertEqual(self.child._captured_events[0][1].buttons, 123)

        self.assertTrue(self.child.has_focus())

    def test_mouse_propagate(self):
        self.parent.event(widgets.Event.mouse_event(geo.Point(1, 2), 123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 1)
        self.assertEqual(len(self.sibling._captured_events), 1)

        self.assertEqual(self.sibling._captured_events[0][0], "mouse_event")
        self.assertEqual(self.sibling._captured_events[0][1].pos, geo.Point(1, 2))
        self.assertEqual(self.sibling._captured_events[0][1].buttons, 123)

    def test_mouse_skip_inactive(self):
        self.child.active = False
        self.parent.event(widgets.Event.mouse_event(geo.Point(1, 2), 123))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(len(self.sibling._captured_events), 1)

    def test_mouse_deny_focus(self):
        self.no_focus()
        self.child_mouse_event.propagate = False
        with patch.object(self.parent, "event_change_focus_child", lambda x: False):
            self.parent.event(widgets.Event.mouse_event(geo.Point(1, 2), 123))
        self.assertFalse(self.child.has_focus())

    def test_resize_nopropagate(self):
        self.resize_event.propagate = False
        rect = geo.Rect(x=0, y=0, width=10, height=10)
        self.parent.event(widgets.Event.resize_event(rect))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 0)
        self.assertEqual(self.parent._captured_events[0][0], "resize_event")
        self.assertEqual(self.parent._captured_events[0][1].bounds, rect)

    def test_resize_propagate(self):
        rect = geo.Rect(x=0, y=0, width=10, height=10)
        self.parent.adjust_window_size(rect)
        self.parent.event(widgets.Event.resize_event(rect))
        self.assertEqual(len(self.parent._captured_events), 1)
        self.assertEqual(len(self.child._captured_events), 1)
        self.assertEqual(self.child._captured_events[0][0], "resize_event")
        self.assertEqual(self.child._captured_events[0][1].bounds, rect)


class CaptureEvent(object):
    def __init__(self, object, name, propagate=True):
        self.object = object
        self.name = name
        self.propagate = propagate
        self.object._captured_events = []
        setattr(self.object, name, self)

    def __call__(self, event):
        self.object._captured_events.append((self.name, event))
        event.propagate = self.propagate


class TestTabBar(test_common.TestCase):
    def setUp(self):
        self.widget = widgets.TabBar(
            None,
            MockCursesWindow(1, 64),
            ["foo", "bar", "baz"],
            None,
        )
        self.changed = []
        self.widget.signal_changed = self.changed.append

    def test_key_event(self):
        self.widget.window.input_queue = [
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_LEFT,
            curses.KEY_LEFT,
            curses.KEY_HOME,
            curses.KEY_END,
            curses.KEY_F1,
            ord("q"),
        ]

        with patch.object(self.widget, "text_event", lambda e: self.widget.deactivate()):
            self.widget.loop()
        self.assertEqual(self.changed, ["foo", "bar", "baz", "foo", "baz", "bar", "foo", "baz"])

        self.widget.items = []
        self.widget.current = "foo"
        self.widget.key_event(widgets.Event.key_event(curses.KEY_RIGHT))
        self.assertIsNone(self.widget.current)
        self.widget.current = "foo"
        self.widget.key_event(widgets.Event.key_event(curses.KEY_HOME))
        self.assertIsNone(self.widget.current)
        self.widget.current = "foo"
        self.widget.key_event(widgets.Event.key_event(curses.KEY_END))
        self.assertIsNone(self.widget.current)

    def test_window_bounds_hint(self):
        rect = geo.Rect(x=4, y=7, width=25, height=84)
        self.assertEqual(
            self.widget.window_bounds_hint(rect),
            geo.Rect(x=rect.x, y=rect.y, width=rect.width, height=1)
        )

    def test_mouse_event(self):
        event = widgets.Event.mouse_event(geo.Point(5, 0), curses.BUTTON2_CLICKED)
        self.widget.mouse_event(event)
        self.assertTrue(event.propagate)

        event.buttons = curses.BUTTON1_CLICKED
        self.widget.mouse_event(event)
        self.assertFalse(event.propagate)
        self.assertEqual(self.widget.current, "bar")

        event = widgets.Event.mouse_event(geo.Point(99, 0), curses.BUTTON1_CLICKED)
        self.widget.mouse_event(event)
        self.assertFalse(event.propagate)
        self.assertEqual(self.widget.current, "bar")

    def test_no_extra_switch(self):
        self.widget.window.input_queue = [
            curses.KEY_RIGHT,
            curses.KEY_HOME,
            ord("q"),
        ]
        with patch.object(self.widget, "text_event", lambda e: self.widget.deactivate()):
            self.widget.loop()
        self.assertEqual(self.changed, ["foo"])

    def test_default_action(self):
        self.widget = widgets.TabBar(
            None,
            MockCursesWindow(1, 64),
            ["foo", "bar", "baz"],
            None,
        )
        self.widget.window.input_queue = [
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_RIGHT,
            curses.KEY_LEFT,
            curses.KEY_LEFT,
            curses.KEY_HOME,
            curses.KEY_END,
            ord("q"),
        ]
        self.widget.text_event = lambda *args: self.widget.deactivate()
        self.widget.loop()

    def test_render(self):
        self.widget.current = self.widget.items[0]

        separator_string = (self.widget.separator, curses.A_NORMAL)
        strings = self.widget.window.strings

        self.widget.refresh()
        self.assertEqual(strings[(0, 0)], separator_string)
        self.assertEqual(strings[(1, 0)], ("foo", curses.A_REVERSE))
        self.assertEqual(strings[(4, 0)], separator_string)
        self.assertEqual(strings[(5, 0)], ("bar", curses.A_NORMAL))
        self.assertEqual(strings[(8, 0)], separator_string)
        self.assertEqual(strings[(9, 0)], ("baz", curses.A_NORMAL))
        self.assertEqual(strings[(12, 0)], separator_string)

        with patch.object(self.widget, "has_focus", new=lambda: False):
            self.widget.refresh()
            self.assertEqual(strings[(0, 0)], separator_string)
            self.assertEqual(strings[(1, 0)], ("foo", curses.A_UNDERLINE))
            self.assertEqual(strings[(4, 0)], separator_string)
            self.assertEqual(strings[(5, 0)], ("bar", curses.A_NORMAL))
            self.assertEqual(strings[(8, 0)], separator_string)
            self.assertEqual(strings[(9, 0)], ("baz", curses.A_NORMAL))
            self.assertEqual(strings[(12, 0)], separator_string)


class TestMenu(test_common.TestCase):
    class Item(widgets.Menu.Item):
        def __init__(self, name, test_obj):
            super(TestMenu.Item, self).__init__(name)
            self.action = lambda: test_obj.activated.append(name)

    def setUp(self):
        self.activated = []
        self.widget = widgets.Menu(
            None,
            MockCursesWindow(24, 80),
            [
                TestMenu.Item("foo", self),
                TestMenu.Item("bar", self),
                TestMenu.Item("baz", self),
            ],
        )

    def test_ctor(self):
        self.assertIs(self.widget.current, self.widget.items[0])

        other = widgets.Menu(None, MockCursesWindow(24, 80))
        self.assertIsNone(other.current)

    def test_window_bounds_hint(self):
        self.assertEqual(
            self.widget.window_bounds_hint(geo.Rect(x1=1, y1=2, x2=116, y2=92)),
            geo.Rect(x=100, y=2, width=16, height=90),
        )

    def test_key_events(self):
        self.widget.window.input_queue = [
            curses.KEY_DOWN,
            curses.KEY_ENTER,
            curses.KEY_DOWN,
            curses.KEY_ENTER,
            ord(" "),
            curses.KEY_DOWN,
            curses.KEY_ENTER,
            curses.KEY_DOWN,
            curses.KEY_ENTER,
            curses.KEY_UP,
            curses.KEY_UP,
            ord(" "),
            curses.KEY_HOME,
            ord(" "),
            curses.KEY_END,
            ord(" "),
            curses.KEY_LEFT,
            ord("f"),
            ord("q"),
        ]

        old_te = self.widget.text_event
        def new_te(ev):
            if ev.char == "q":
                self.widget.deactivate()
            else:
                old_te(ev)

        self.widget.current = None
        with patch.object(self.widget, "text_event", new_te):
            self.widget.loop()

        self.assertEqual(
            self.activated,
            ["foo", "bar", "bar", "baz", "foo", "bar", "foo", "baz"]
        )

    def test_key_events_noitems(self):
        self.widget.items = []

        self.widget.current = 1
        self.widget.key_event(widgets.Event.key_event(curses.KEY_HOME))
        self.assertIsNone(self.widget.current)

        self.widget.current = 1
        self.widget.key_event(widgets.Event.key_event(curses.KEY_END))
        self.assertIsNone(self.widget.current)

        self.widget.current = 1
        self.widget.key_event(widgets.Event.key_event(curses.KEY_UP))
        self.assertIsNone(self.widget.current)

        self.widget.current = 1
        self.widget.key_event(widgets.Event.key_event(curses.KEY_DOWN))
        self.assertIsNone(self.widget.current)

    def test_activate(self):
        self.widget.activate()
        self.assertEqual(self.activated, ["foo"])

        self.activated = []
        self.widget.current = None
        self.widget.activate()
        self.assertEqual(self.activated, [])

    def test_submenu(self):
        sub1 = [
            TestMenu.Item("a1", self),
            TestMenu.Item("a2", self),
            TestMenu.Item("a3", self),
        ]
        sub2 = [
            TestMenu.Item("b1", self),
            TestMenu.Item("b2", self),
            TestMenu.Item("b3", self),
        ]
        sub3 = []

        self.widget.push_submenu(sub1)
        self.assertIn(self.widget.current, sub1)
        self.assertEqual(self.widget.items, sub1)

        self.widget.push_submenu(sub2)
        self.assertIn(self.widget.current, sub2)
        self.assertEqual(self.widget.items, sub2)

        self.widget.push_submenu(sub3)
        self.assertIsNone(self.widget.current)
        self.assertEqual(self.widget.items, [])

        self.widget.pop_submenu()
        self.assertIn(self.widget.current, sub2)
        self.assertEqual(self.widget.items, sub2)

        self.widget.key_event(widgets.Event.key_event(curses.KEY_BACKSPACE))
        self.assertIn(self.widget.current, sub1)
        self.assertEqual(self.widget.items, sub1)

        self.widget.key_event(widgets.Event.key_event(curses.KEY_BACKSPACE))
        self.widget.key_event(widgets.Event.key_event(curses.KEY_BACKSPACE))
        self.assertNotEqual(self.widget.items, sub1)
        self.assertEqual(len(self.widget.items), 3)

        self.assertRaises(OverflowError, self.widget.pop_submenu)

    def test_focus_event(self):
        self.widget.current = None

        self.widget.focus_event(widgets.Event.focus_event(False))
        self.assertIsNone(self.widget.current)

        self.widget.focus_event(widgets.Event.focus_event(True))
        self.assertIn(self.widget.current, self.widget.items)

        self.widget.focus_event(widgets.Event.focus_event(False))
        self.assertIn(self.widget.current, self.widget.items)

        self.widget.current = None
        self.widget.items = []
        self.widget.focus_event(widgets.Event.focus_event(True))
        self.assertIsNone(self.widget.current)

    def test_auto_activate(self):
        self.widget.items[0].auto_activate = True
        self.widget.items[2].auto_activate = True
        self.widget.current = None


        self.widget.window.input_queue = [
            curses.KEY_DOWN,
            curses.KEY_DOWN,
            curses.KEY_DOWN,
            ord("q")
        ]

        self.widget.current = None
        with patch.object(self.widget, "text_event",
                          lambda x: self.widget.deactivate()):
            self.widget.loop()

        self.assertEqual(self.activated, ["foo", "baz"])

    def test__starting_y(self):
        with patch.object(self.widget, "window_bounds",
                return_value=geo.Rect(x=100, y=100, width=10, height=3)):
            self.assertEqual(self.widget._starting_y(), 100)

        with patch.object(self.widget, "window_bounds",
                return_value=geo.Rect(x=100, y=200, width=10, height=5)):
            self.assertEqual(self.widget._starting_y(), 201)

        with patch.object(self.widget, "window_bounds",
                return_value=geo.Rect(x=0, y=0, width=80, height=24)):
            self.assertEqual(self.widget._starting_y(), int((24-3)/2))

    def test_render(self):
        self.widget.current = self.widget.items[0]

        strings = self.widget.window.strings

        self.widget.refresh()
        starting_y = self.widget._starting_y()
        self.assertEqual(strings[(1, starting_y)], ("foo", curses.A_REVERSE))
        self.assertEqual(strings[(1, starting_y+1)], ("bar", curses.A_NORMAL))
        self.assertEqual(strings[(1, starting_y+2)], ("baz", curses.A_NORMAL))

        with patch.object(self.widget, "has_focus", new=lambda: False):
            self.widget.refresh()
            self.assertEqual(strings[(1, starting_y)], ("foo", curses.A_UNDERLINE))
            self.assertEqual(strings[(1, starting_y+1)], ("bar", curses.A_NORMAL))
            self.assertEqual(strings[(1, starting_y+2)], ("baz", curses.A_NORMAL))

    def test_mouse_event(self):
        self.widget.current = None
        ev = widgets.Event.mouse_event(
            geo.Point(1, self.widget._starting_y() + 1),
            curses.BUTTON1_CLICKED
        )
        self.widget.mouse_event(ev)
        self.assertIs(self.widget.current, self.widget.items[1])
        self.assertFalse(ev.propagate)

        ev = widgets.Event.mouse_event(
            geo.Point(1, 0),
            curses.BUTTON1_CLICKED
        )
        self.widget.mouse_event(ev)
        self.assertIs(self.widget.current, self.widget.items[1])
        self.assertFalse(ev.propagate)

        self.widget.current = None
        ev = widgets.Event.mouse_event(
            geo.Point(1, self.widget._starting_y() + 1),
            curses.BUTTON2_CLICKED
        )
        self.widget.mouse_event(ev)
        self.assertIsNone(self.widget.current)
        self.assertTrue(ev.propagate)

    def test_default_action(self):
        item = widgets.Menu.Item("hello")
        self.assertTrue(callable(item.action))
        item.action()


class TestTabWidget(test_common.TestCase):
    def setUp(self):
        self.widget = widgets.TabWidget(
            None,
            MockCursesWindow(24, 80),
            lambda wid: getattr(wid, "name", "Unnamed Widget")
        )

    def _create_tab(self, name):
        widget = self.widget.create_tab(widgets.Widget)
        widget.name = name
        return widget

    def _create_tabs(self, names=["foo", "bar", "baz"]):
        return [self._create_tab(name) for name in names]


    def test_container_area(self):
        self.assertEqual(
            self.widget.container.window_bounds(),
            geo.Rect(x=0, y=1, width=80, height=23)
        )

    def test_create_tab_noadd(self):
        widget = self.widget.create_tab_noadd()
        self.assertIsInstance(widget, widgets.Widget)
        self.assertNotIn(widget, self.widget.tabs)
        self.assertEqual(len(self.widget.tabs), 0)

    def test_create_tab(self):
        widget = self.widget.create_tab()
        self.assertIsInstance(widget, widgets.Widget)
        self.assertEqual(len(self.widget.tabs), 1)
        self.assertIn(widget, self.widget.tabs)
        self.assertIs(widget, self.widget.current_tab)
        self.assertIs(widget, self.widget.bar.current)

    def test_add_tab(self):
        widget = self.widget.create_tab_noadd()
        self.assertIsInstance(widget, widgets.Widget)
        self.assertNotIn(widget, self.widget.tabs)

        self.widget.add_tab(widget)
        self.assertEqual(len(self.widget.tabs), 1)
        self.assertIn(widget, self.widget.tabs)
        self.assertIs(widget, self.widget.current_tab)
        self.assertIs(widget, self.widget.bar.current)

        self._create_tab("foo")
        self.widget.add_tab(widget)
        self.widget.add_tab(widget)
        self.assertEqual(len(self.widget.tabs), 2)

    def test__switch_current(self):
        tabs = self._create_tabs()
        self.assertEqual(tabs, self.widget.tabs)
        self.widget._switch_current(tabs[1])
        self.assertIs(tabs[1], self.widget.current_tab)
        self.assertIs(tabs[1], self.widget.bar.current)
        self.assertFalse(tabs[0].active)
        self.assertTrue(tabs[1].active)
        self.assertFalse(tabs[2].active)
        self.assertTrue(self.widget.window.refreshed)

        self.widget.window.refreshed = False
        self.widget._switch_current(tabs[1])
        self.assertFalse(self.widget.window.refreshed)

        self.widget._switch_current(None)
        self.assertIsNone(self.widget.current_tab)
        self.assertIsNone(self.widget.bar.current)

        self.widget._switch_current(tabs[1])
        self.assertIs(tabs[1], self.widget.current_tab)
        self.assertIs(tabs[1], self.widget.bar.current)

    def test_close_current(self):
        tabs = self._create_tabs()
        self.widget._switch_current(tabs[1])
        self.widget.close_current()
        self.assertIs(tabs[2], self.widget.current_tab)
        self.assertEqual(len(self.widget.tabs), 2)

        self.widget.current_tab = None
        self.widget.close_current()
        self.assertEqual(len(self.widget.tabs), 2)

        self.widget.current_tab = tabs[2]
        self.widget.close_current()
        self.assertIs(tabs[0], self.widget.current_tab)
        self.assertEqual(len(self.widget.tabs), 1)

        self.widget.close_current()
        self.assertIsNone(self.widget.current_tab)
        self.assertEqual(len(self.widget.tabs), 0)

        self.widget.close_current()

    def test_formatter(self):
        tabs = self._create_tabs()
        self.assertEqual(self.widget.bar.formatter(tabs[0]), "foo")

        self.widget = widgets.TabWidget(
            None,
            MockCursesWindow(24, 80),
        )
        tabs = self._create_tabs()
        self.assertEqual(self.widget.bar.formatter(tabs[0]), "tab")

    def test_on_activated(self):
        self.widget._activated = None

        def on_activated(tab):
            self.widget._activated = tab

        with patch.object(self.widget, "on_activated", on_activated):
            tabs = self._create_tabs()
            self.widget._activated = None
            self.widget.activate_tab()
            self.assertIs(self.widget._activated, tabs[2])

            self.widget._switch_current(None)
            self.widget.activate_tab()
            self.assertIs(self.widget._activated, tabs[2])


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import cairosvg
import six

from .. import tree
from .svg import SvgFormatter
from . import factory


class PngFormatter(object):
    flat = True

    def __init__(self, *args, **kwargs):
        self._svg_formatter = SvgFormatter(*args, **kwargs)

    def document(self, doc, output):
        buffer = six.StringIO()
        self._svg_formatter.document(doc, buffer)
        cairosvg.svg2png(bytestring=buffer.getvalue(), write_to=output)

    def layer(self, layer, output):
        doc = tree.Document()
        doc.layers.append(layer)
        self.document(doc, output)

    def color(self, color):
        return ""


factory.register(PngFormatter(), "png")



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from .widget import Widget
from .geometry import Point, Rect
from ._util import next_object
import curses


class Menu(Widget):
    class Item(object):
        def __init__(self, text, action=lambda: None, display_flags=0, auto_activate=False):
            self.text = text
            self.action = action
            self.display_flags = display_flags
            self.auto_activate = auto_activate

    def __init__(self, parent, window, items=[], width=16):
        self.width = width
        super(Menu, self).__init__(parent, window)
        self.items = items
        self.current = items[0] if items else None
        self.submenu = []

    def window_bounds_hint(self, parent_bounds):
        return Rect(
            top_left=Point(parent_bounds.right - self.width, parent_bounds.top),
            bottom_right=parent_bounds.bottom_right
        )

    def key_event(self, event):
        if event.key == curses.KEY_UP:
            self._switch_item(-1)
        elif event.key == curses.KEY_DOWN:
            self._switch_item(+1)
        elif event.key == curses.KEY_ENTER:
            self.activate()
        elif event.key == curses.KEY_BACKSPACE:
            if self.submenu:
                self.pop_submenu()
        elif event.key == curses.KEY_HOME:
            if not self.items:
                self._change_current(None)
            else:
                self._change_current(self.items[0])
        elif event.key == curses.KEY_END:
            if not self.items:
                self._change_current(None)
            else:
                self._change_current(self.items[-1])

    def text_event(self, event):
        if event.char == " ":
            self.activate()

    def _switch_item(self, delta):
        if not self.items:
            self._change_current(None)
        elif not self.current:
            self._change_current(self.items[0])
        else:
            self._change_current(next_object(self.items, self.current, delta))

    def _change_current(self, item):
        self.current = item

        if self.current is not None and self.current.auto_activate:
            self.current.action()

        self.refresh()

    def activate(self):
        if self.current:
            self.current.action()

    def _starting_y(self):
        return int(
            self.window_bounds().top +
            (self.window_bounds().height - len(self.items)) / 2
        )

    def render(self):
        y = self._starting_y()
        for item in self.items:
            if item is not self.current:
                mode = 0
            elif self.has_focus():
                mode = curses.A_REVERSE
            else:
                mode = curses.A_UNDERLINE
            self.window.addstr(y, 1, item.text, mode|item.display_flags)
            y += 1
        # TODO handle overflow

        self._draw_border()

    def focus_event(self, event):
        if event.focus and not self.current and self.items:
            self.current = self.items[0]
        super(Menu, self).focus_event(event)

    def push_submenu(self, items):
        self.submenu.append((self.items, self.current))
        self.items = items
        self.current = items[0] if items else None
        self.refresh()

    def pop_submenu(self):
        if not self.submenu:
            raise OverflowError("Underflow")
        cur_items = self.items
        self.items, self.current = self.submenu.pop()
        self.refresh()
        return cur_items

    def mouse_event(self, event):
        super(Menu, self).mouse_event(event)
        if event.buttons == curses.BUTTON1_CLICKED:
            index = event.pos.y - self._starting_y()
            if 0 <= index < len(self.items):
                self._change_current(self.items[index])
                self.activate()
                self.refresh()
            event.accept()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import six
import json
from xml.etree import ElementTree
from mock import patch
from contextlib import contextmanager

import test_common

from patsi import ansi
from patsi.document import tree
from patsi.document import color
from patsi.document import palette
from patsi.document import formatter
from patsi.document.formatter import factory


class TestFactory(test_common.StringOutputTestCase):
    formatters = {
        "ansi": formatter.AnsiFormatter,
        "sh": formatter.AnsiSourceFormatter,
        "py": formatter.AnsiSourceFormatter,
        "pl": formatter.AnsiSourceFormatter,
        "php": formatter.AnsiSourceFormatter,
        "irc": formatter.IrcFormatter,
        "json": formatter.JsonFormatter,
        "png": formatter.PngFormatter,
        "svg": formatter.SvgFormatter,
        "txt": formatter.TextFormatter,
        "xml": formatter.XmlFormatter,
    }

    def test_formatter(self):
        for name, cls in six.iteritems(TestFactory.formatters):
            self.assertIs(type(factory.formatter(name)), cls)
        self.assertRaises(KeyError, lambda: factory.formatter("____"))

    def test_formats(self):
        self.assertSetEqual(set(TestFactory.formatters), set(factory.formats()))

    def test_formatter_for_file(self):
        for name, cls in six.iteritems(TestFactory.formatters):
            self.assertIs(type(factory.formatter_for_file("foo." + name)), cls)
        self.assertRaises(KeyError, lambda: factory.formatter_for_file("foo.bar"))

    def test_save(self):
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
        )

        self.output.name = "foo.____"
        self.assertRaises(KeyError, lambda: factory.save(doc, self.output))

        self.output.name = "foo.txt"
        self._clear_data()
        factory.save(doc, self.output, "ansi")
        self._check_data(ansi.SGR(ansi.SGR.Red), "Hello\n",
                         ansi.SGR(ansi.SGR.Blue), "World\n")

        self._clear_data()
        factory.save(doc, self.output)
        self._check_data("Hello\nWorld\n")

        self._clear_data()
        mock_file = test_common.MockFile(self.output)
        with patch("patsi.document.formatter.factory.open", mock_file.open):
            factory.save(doc, "foo/bar.txt")
            self._check_data("Hello\nWorld\n")

        self._clear_data()
        factory.save(doc.layers[0], self.output)
        self._check_data("Hello\n")


class TestAnsiFormatter(test_common.StringOutputTestCase):
    fmt = formatter.AnsiFormatter()

    def test_flat(self):
        self.assertTrue(self.fmt.flat)

    def test_color(self):
        colors = [
            (None, ansi.SGR.ResetColor),
            (color.RgbColor(1, 2, 3), ansi.SGR.ColorRGB(1, 2, 3)),
            (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
            (color.IndexedColor(1, palette.colors8_bright),
                ansi.SGR.Color(ansi.SGR.Color.Red, bright=True)),
            (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
            (color.IndexedColor(9, palette.colors16),
                ansi.SGR.Color(ansi.SGR.Color.Red, bright=True)),
            (color.IndexedColor(9, palette.colors256), ansi.SGR.Color256(9)),
        ]

        for doc_col, sgr_col in colors:
            self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))

        self.assertRaises(TypeError, lambda: self.fmt.color(69))

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self._check_data(ansi.SGR(ansi.SGR.Red), layer.text)
        self._clear_data()

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self.fmt.layer(layer, self.output)
        self._check_data(ansi.SGR(ansi.SGR.Red), "Hell", ansi.SGR(ansi.SGR.Blue), "o!\n")
        self._clear_data()

        layer = tree.FreeColorLayer()
        layer.set_char(1, 0, "H")
        layer.set_char(2, 0, "i")
        self.fmt.layer(layer, self.output)
        self._check_data(" Hi\n")
        self._clear_data()

        layer = None
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")

    def test_layer_newline_color(self):
        red = color.IndexedColor(1, palette.colors16)
        sgr_red = str(ansi.SGR(ansi.SGR.Red))

        layer = tree.Layer("Hello\nWorld", red)
        self.fmt.layer(layer, self.output)
        self.assertTrue(all(
            line.startswith(sgr_red)
            for line in self._get_data().splitlines()
        ))
        self._clear_data()

        layer = tree.FreeColorLayer()
        self.fmt.layer(layer, self.output)
        layer.set_char(0, 0, "H", red)
        layer.set_char(0, 1, "i", red)
        self.assertTrue(all(
            line.startswith(sgr_red)
            for line in self._get_data().splitlines()
        ))
        self._clear_data()

    def test_document(self):
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
        )
        self.fmt.document(doc, self.output)
        self._check_data(ansi.SGR(ansi.SGR.Red), "Hello\n",
                         ansi.SGR(ansi.SGR.Blue), "World\n")

    def test_features(self):
        Features = formatter.AnsiFormatter.Features
        self.assertEqual(Features.feature_from_name("standard"), Features.STANDARD)
        self.assertEqual(Features.feature_from_name("Bright_Bold"), Features.BRIGHT_BOLD)
        self.assertIn("EXT_PALETTE", Features.constants_dict())

        feat = Features(69)
        self.assertEqual(feat.features, 69)
        self.assertEqual(Features(feat).features, 69)
        self.assertIsNot(Features(feat), feat)
        self.assertEqual(Features().features, Features.ALL)

        self.assertEqual(Features("Bright_Bold").features, Features.BRIGHT_BOLD)
        self.assertEqual(Features("Bright_Bold, RGB").features, Features.BRIGHT_BOLD|Features.RGB)
        self.assertEqual(Features(["Bright_Bold", "RGB"]).features, Features.BRIGHT_BOLD|Features.RGB)
        self.assertEqual(Features([Features.BRIGHT_BOLD, Features.RGB]).features, Features.BRIGHT_BOLD|Features.RGB)

        self.assertTrue(Features(0).has_feature(0))
        self.assertTrue(Features(3).has_feature(2))
        self.assertFalse(Features(0).has_feature(1))
        self.assertFalse(Features(3).has_feature(4))
        self.assertTrue(Features(Features.BRIGHT_BOLD).has_feature("Bright_Bold"))

    @contextmanager
    def features(self, features):
        self.fmt.features = formatter.AnsiFormatter.Features(features)
        try:
            yield
        finally:
           self.fmt.features = formatter.AnsiFormatter.Features()

    def test_colors_standard(self):
        with self.features("standard"):
            colors = [
                (None, ansi.SGR.ResetColor),
                (color.RgbColor(255, 2, 3), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors8_bright), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors256), ansi.SGR.Red),
            ]

            for doc_col, sgr_col in colors:
                self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))

    def test_colors_bright(self):
        with self.features("bright"):
            bright_red = ansi.SGR.Color(ansi.SGR.Color.Red, bright=True)
            colors = [
                (None, ansi.SGR.ResetColor),
                (color.RgbColor(255, 2, 3), bright_red),
                (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors8_bright), bright_red),
                (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors16), bright_red),
                (color.IndexedColor(9, palette.colors256), bright_red),
            ]

            for doc_col, sgr_col in colors:
                self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))

    def test_colors_256(self):
        with self.features("EXT_PALETTE"):
            bright_red = ansi.SGR.Color256(9)
            colors = [
                (None, ansi.SGR.ResetColor),
                (color.RgbColor(255, 2, 3), bright_red),
                (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors8_bright), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors256), bright_red),
            ]

            for doc_col, sgr_col in colors:
                self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))

    def test_colors_bright_bold(self):
        with self.features("bright_bold"):
            bright_red = [ansi.SGR.Red, ansi.SGR.Bold]
            dark_red = [ansi.SGR.Red, ansi.SGR.reverse(ansi.SGR.Bold)]
            colors = [
                (None, ansi.SGR.ResetColor),
                (color.RgbColor(255, 2, 3), bright_red),
                (color.IndexedColor(1, palette.colors8_dark), dark_red),
                (color.IndexedColor(1, palette.colors8_bright), bright_red),
                (color.IndexedColor(1, palette.colors16), dark_red),
                (color.IndexedColor(9, palette.colors16), bright_red),
                (color.IndexedColor(9, palette.colors256), bright_red),
            ]

            for doc_col, sgr_col in colors:
                self.fmt._bold = True
                self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))

    def test_colors_rgb(self):
        with self.features("rgb"):
            colors = [
                (None, ansi.SGR.ResetColor),
                (color.RgbColor(255, 2, 3), ansi.SGR.ColorRGB(255, 2, 3)),
                (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors8_bright), ansi.SGR.Red),
                (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors16), ansi.SGR.Red),
                (color.IndexedColor(9, palette.colors256), ansi.SGR.ColorRGB(255, 0, 0)),
            ]

            for doc_col, sgr_col in colors:
                self.assertEqual(self.fmt.color(doc_col), repr(ansi.SGR(sgr_col)))


class TestAnsiSourceFormatter(test_common.StringOutputTestCase):
    fmt = formatter.AnsiSourceFormatter()

    def test_flat(self):
        self.assertTrue(self.fmt.flat)

    def test_color(self):
        self.assertEqual(
            self.fmt.color(color.IndexedColor(1, palette.colors8_dark)),
            r'"\x1b[31m"'+"\n"
        )

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self._check_data(r'"\x1b[31mHello World\n"', "\n")
        self._clear_data()

    def test_document(self):
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
        )
        self.fmt.document(doc, self.output)
        self._check_data(r'"\x1b[31mHello\n\x1b[34mWorld\n"', "\n")


class TestTextFormatter(test_common.StringOutputTestCase):
    fmt = formatter.TextFormatter()

    def test_flat(self):
        self.assertTrue(self.fmt.flat)

    def test_color(self):
        colors = [
            (None, ansi.SGR.ResetColor),
            (color.RgbColor(1, 2, 3), ansi.SGR.ColorRGB(1, 2, 3)),
            (color.IndexedColor(1, palette.colors8_dark), ansi.SGR.Red),
            (color.IndexedColor(1, palette.colors8_bright),
                ansi.SGR.Color(ansi.SGR.Color.Red, bright=True)),
            (color.IndexedColor(1, palette.colors16), ansi.SGR.Red),
            (color.IndexedColor(9, palette.colors16),
                ansi.SGR.Color(ansi.SGR.Color.Red, bright=True)),
            (color.IndexedColor(9, palette.colors256), ansi.SGR.Color256(9)),
        ]

        for doc_col, sgr_col in colors:
            self.assertEqual(self.fmt.color(doc_col), "")

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self._check_data(layer.text)
        self._clear_data()

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self.fmt.layer(layer, self.output)
        self._check_data("Hello!\n")
        self._clear_data()

        layer = None
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")

    def test_document(self):
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
        )
        self.fmt.document(doc, self.output)
        self._check_data("Hello\nWorld\n")


class TestIrcFormatter(test_common.StringOutputTestCase):
    fmt = formatter.IrcFormatter()
    irc_colors = formatter.IrcFormatter.colors

    def test_flat(self):
        self.assertTrue(self.fmt.flat)

    def test_color(self):

        colors = [
            (color.RgbColor(1, 2, 155), 8+4),
            (color.IndexedColor(1, palette.colors8_dark), 1),
            (color.IndexedColor(1, palette.colors8_bright), 8+1),
            (color.IndexedColor(1, palette.colors16), 1),
            (color.IndexedColor(9, palette.colors16), 8+1),
            (color.IndexedColor(9, palette.colors256), 8+1),
        ]

        for doc_col, ansi_col in colors:
            irc_col = self.irc_colors[ansi_col]
            result = self.fmt.color(doc_col)
            self.assertEqual(result, "\x03%s,01" % irc_col,
                "converting %s to %s (%s) resulted in %s" %
                    (doc_col.name, irc_col, ansi_col, result)
            )

        self.assertEqual(self.fmt.color(None), "\x0f")

        self.assertRaises(TypeError, lambda: self.fmt.color(69))

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self._check_data(self.fmt.color(red), layer.text)

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        self._check_data(self.fmt.color(red), "Hell",
                         self.fmt.color(blue), "o!\n")

        layer = None
        self._clear_data()
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")

        layer = tree.FreeColorLayer()
        layer.set_char(1, 0, "H")
        layer.set_char(2, 0, "i")
        self.fmt.layer(layer, self.output)
        self._check_data(" Hi\n")
        self._clear_data()

    def test_layer_newline_color(self):
        red = color.IndexedColor(1, palette.colors16)
        out_red = self.fmt.color(red)

        layer = tree.Layer("Hello\nWorld", red)
        self.fmt.layer(layer, self.output)
        self.assertTrue(all(
            line.startswith(out_red)
            for line in self._get_data().splitlines()
        ))
        self._clear_data()

        layer = tree.FreeColorLayer()
        self.fmt.layer(layer, self.output)
        layer.set_char(0, 0, "H", red)
        layer.set_char(0, 1, "i", red)
        self.assertTrue(all(
            line.startswith(out_red)
            for line in self._get_data().splitlines()
        ))
        self._clear_data()

    def test_document(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)

        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", red),
                tree.Layer("\nWorld", blue),
            ],
        )
        self.fmt.document(doc, self.output)
        self._check_data(self.fmt.color(red), "Hello\n",
                         self.fmt.color(blue), "World\n")


class TestJsonFormatter(test_common.StringOutputTestCase):
    fmt = formatter.JsonFormatter()
    maxDiff = None

    def test_flat(self):
        self.assertFalse(self.fmt.flat)

    def test_color(self):
        colors = [
            (None, ""),
            (color.RgbColor(1, 2, 3), "#010203"),
            (color.IndexedColor(1, palette.colors8_dark), "red"),
            (color.IndexedColor(1, palette.colors8_bright), "red"),
            (color.IndexedColor(1, palette.colors16), "red"),
            (color.IndexedColor(9, palette.colors16), "red_bright"),
            (color.IndexedColor(9, palette.colors256), palette.colors256.name(9)),
        ]

        for doc_col, name in colors:
            self.assertEqual(self.fmt.color(doc_col), '"%s"' % name)

        self.assertRaises(Exception, lambda: self.fmt.color(69))

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self.assertDictEqual(
            json.loads(self._get_data()),
            {"color": red.name, "text": layer.text}
        )
        self._clear_data()

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self.fmt.layer(layer, self.output)
        self.assertDictEqual(
            json.loads(self._get_data()),
            {"chars": [
                {"x": 0, "y": 0, "char": "H", "color": "red"},
                {"x": 1, "y": 0, "char": "e", "color": "red"},
                {"x": 2, "y": 0, "char": "l", "color": "red"},
                {"x": 3, "y": 0, "char": "l", "color": "red"},
                {"x": 4, "y": 0, "char": "o", "color": "blue"},
                {"x": 5, "y": 0, "char": "!", "color": "blue"},
            ]}
        )
        self._clear_data()

        layer = None
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")

    def test_document(self):
        metadata = {"hello": "world"}
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
            metadata
        )
        self.fmt.document(doc, self.output)
        self.assertDictEqual(
            json.loads(self._get_data()),
            {
                "layers": [
                    {"color": "red", "text": "Hello\n"},
                    {"color": "blue", "text": "\nWorld\n"},
                ],
                "name": "test",
                "metadata": metadata,
            }
        )


class TestPngFormatter(test_common.BytesOutputTestCase):
    fmt = formatter.PngFormatter()

    def test_flat(self):
        self.assertTrue(self.fmt.flat)

    def test_color(self):
        colors = [
            None,
            color.RgbColor(1, 2, 255),
            color.IndexedColor(1, palette.colors8_dark),
            color.IndexedColor(1, palette.colors8_bright),
            color.IndexedColor(1, palette.colors16),
            color.IndexedColor(9, palette.colors16),
            color.IndexedColor(9, palette.colors256),
        ]

        for doc_col in colors:
            self.assertEqual(self.fmt.color(doc_col), "")

    def _check_png_header(self):
        return self._get_data().startswith(b"PNG")

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self.fmt.layer(layer, self.output)
        self._check_png_header()

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        self._check_png_header()

        layer = None
        self._clear_data()
        self.assertRaises(Exception, lambda: self.fmt.layer(layer, self.output))
        self._check_data(b"")

    def test_document(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)

        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", red),
                tree.Layer("\nWorld", blue),
            ],
        )
        self.fmt.document(doc, self.output)
        self._check_png_header()


class TestSVGFormatter(test_common.StringOutputTestCase):
    fmt = formatter.SvgFormatter()
    xmlns = "http://www.w3.org/2000/svg"

    def _assert_element(self, elem, name):
        self.assertEqual(elem.tag, "{%s}%s" % (self.xmlns, name))

    def _check_svg(self):
        try:
            elem = ElementTree.fromstring(self._get_data())
            self._assert_element(elem, "svg")
            return elem
        except ElementTree.ParseError as e:
            self.fail("Invalid XML: %s" % e)

    def test_flat(self):
        self.assertFalse(formatter.SvgFormatter().flat)
        self.assertTrue(formatter.SvgFormatter(flatten=True).flat)

    def test_color(self):
        colors = [
            color.RgbColor(1, 2, 255),
            color.IndexedColor(1, palette.colors8_dark),
            color.IndexedColor(1, palette.colors8_bright),
            color.IndexedColor(1, palette.colors16),
            color.IndexedColor(9, palette.colors16),
            color.IndexedColor(9, palette.colors256),
        ]

        for doc_col in colors:
            self.assertEqual(self.fmt.color(doc_col), doc_col.rgb.hex())

        self.assertEqual(self.fmt.color(None), "inherit")

        self.assertRaises(Exception, lambda: self.fmt.color(69))

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._check_svg()
        self.assertEqual(len(elem), 2)
        self._assert_element(elem[0], "rect")
        self._assert_element(elem[1], "text")
        self.assertEqual(
            elem.get("width"),
            str(layer.width * self.fmt.font_width)
        )
        self.assertEqual(
            elem.get("height"),
            str(layer.height * self.fmt.font_height)
        )
        self.assertEqual(
            elem[0].get("width"),
            str(layer.width * self.fmt.font_width)
        )
        self.assertEqual(
            elem[0].get("height"),
            str(layer.height * self.fmt.font_height)
        )
        self._assert_element(elem[1][0], "tspan")
        self.assertEqual(elem[1][0].text.strip(), layer.text.strip())

        layer = tree.Layer("Hello&<World>", red)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._check_svg()
        self.assertEqual(elem[1][0].text.strip(), layer.text.strip())

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._check_svg()
        self.assertEqual(len(elem), 2)
        self._assert_element(elem[0], "rect")
        self._assert_element(elem[1], "text")
        self.assertEqual(
            elem.get("width"),
            str(layer.width * self.fmt.font_width)
        )
        self.assertEqual(
            elem.get("height"),
            str(layer.height * self.fmt.font_height)
        )
        self.assertEqual(
            elem[0].get("width"),
            str(layer.width * self.fmt.font_width)
        )
        self.assertEqual(
            elem[0].get("height"),
            str(layer.height * self.fmt.font_height)
        )
        self.assertEqual(len(elem[1]), len(layer.matrix))
        for x, tspan in enumerate(sorted(elem[1], key=lambda e: float(e.get("x")))):
            self._assert_element(tspan, "tspan")
            self.assertEqual(tspan.text, layer.matrix[(x, 0)][0])
            color_hex = layer.matrix[(x, 0)][1].rgb.hex()
            self.assertIn("fill:%s;" % color_hex, tspan.get("style"))

        layer = None
        self._clear_data()
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")

    def test_document_noflat(self):
        metadata = {"hello": "world"}
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
            metadata
        )
        self.fmt.document(doc, self.output)
        elem = self._check_svg()
        self.assertEqual(len(elem), len(doc.layers)+1)
        self._assert_element(elem[0], "rect")

        for i in range(len(doc.layers)):
            self._assert_element(elem[i+1], "text")
            self.assertEqual(
                elem[i+1][0].text.strip(),
                doc.layers[i].text.strip()
            )

    def test_document_flat(self):
        metadata = {"hello": "world"}
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
            metadata
        )
        self.fmt.flat = True
        self.fmt.document(doc, self.output)
        self.fmt.flat = False
        elem = self._check_svg()
        self.assertEqual(len(elem), 2)
        self._assert_element(elem[0], "rect")
        self._assert_element(elem[1], "text")


class TestXMLFormatter(test_common.BytesOutputTestCase):
    fmt = formatter.XmlFormatter()

    def _get_xml(self):
        try:
            return ElementTree.fromstring(self._get_data())
        except ElementTree.ParseError as e:
            self.fail("Invalid XML: %s" % e)

    def test_color(self):
        colors = [
            color.RgbColor(1, 2, 255),
            color.IndexedColor(1, palette.colors8_dark),
            color.IndexedColor(1, palette.colors8_bright),
            color.IndexedColor(1, palette.colors16),
            color.IndexedColor(9, palette.colors16),
            color.IndexedColor(9, palette.colors256),
        ]

        for doc_col in colors:
            self.assertEqual(self.fmt.color(doc_col), doc_col.name)

        self.assertEqual(self.fmt.color(None), "")

        self.assertRaises(Exception, lambda: self.fmt.color(69))

    def test_layer(self):
        red = color.IndexedColor(1, palette.colors16)
        blue = color.IndexedColor(4, palette.colors16)
        layer = tree.Layer("Hello World", red)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._get_xml()
        self.assertEqual(elem.tag, "layer")
        self.assertEqual(elem.text, layer.text)
        self.assertEqual(elem.get("color"), "red")

        layer = tree.Layer("Hello&<World>", red)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._get_xml()
        self.assertEqual(elem.text, layer.text)

        layer = tree.FreeColorLayer()
        layer.set_char(0, 0, "H", red)
        layer.set_char(1, 0, "e", red)
        layer.set_char(2, 0, "l", red)
        layer.set_char(3, 0, "l", red)
        layer.set_char(4, 0, "o", blue)
        layer.set_char(5, 0, "!", blue)
        self._clear_data()
        self.fmt.layer(layer, self.output)
        elem = self._get_xml()
        self.assertEqual(elem.tag, "layer")
        self.assertEqual(len(elem), len(layer.matrix))

        for x, char in enumerate(sorted(elem, key=lambda e: float(e.get("x")))):
            self.assertEqual(char.tag, "char")
            self.assertEqual(char.text, layer.matrix[(x, 0)][0])
            self.assertEqual(char.get("color"), layer.matrix[(x, 0)][1].name)

        layer = None
        self._clear_data()
        self.assertRaises(TypeError, lambda: self.fmt.layer(layer, self.output))
        self._check_data("")


    def test_document(self):
        doc = tree.Document()
        self.fmt.document(doc, self.output)
        elem = self._get_xml()
        self.assertEqual(len(elem), 0)
        self.assertEqual(elem.tag, "document")

        metadata = {"hello": "world"}
        doc = tree.Document(
            "test",
            [
                tree.Layer("Hello", color.IndexedColor(1, palette.colors16)),
                tree.Layer("\nWorld", color.IndexedColor(4, palette.colors16)),
            ],
            metadata
        )
        self._clear_data()
        self.fmt.document(doc, self.output)
        elem = self._get_xml()
        self.assertEqual(len(elem), len(doc.layers)+1)
        self.assertEqual(elem.tag, "document")
        self.assertEqual(elem.get("name"), "test")

        for i, layer in enumerate(elem.findall("layer")):
            self.assertEqual(layer.text, doc.layers[i].text)

        self.assertEqual(elem.find("metadata").find("hello").text, "world")


test_common.main()



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import re
import sys
import six
import copy
from contextlib import contextmanager


class AnsiCode(object):
    """
    Class representing an ANSI code.

    This class supports a subset of the codes with the
    Control Sequence Introducer
    """

    CSI = "\x1b["
    CSI_PATTERN = re.compile("\x1b\[([<-?]?)([0-9;]*)([@-~])")

    def __init__(self, terminator, args=[], private=""):
        self.terminator = terminator
        self._args = args
        self.private = private

    def __repr__(self):
        """
        Returns the ANSI-encoded string for this code
        """
        return (AnsiCode.CSI + self.private +
                ";".join(str(arg) for arg in self.args()) +
                self.terminator)

    def args(self):
        """
        List of arguments to the ANSI code, as a function to allow overriding
        """
        return self._args

    @staticmethod
    def parse(string):
        """
        Parses an ANSI code string into an object
        """
        match = AnsiCode.CSI_PATTERN.match(string)
        if not match:
            raise Exception(
                "Not a valid escape sequence: %s" %
                string.replace("\x1b", "\\e")
            )
        return AnsiCode._from_match(match)

    @staticmethod
    def _from_match(match):
        """
        Converts a regex match (matched from CSI_PATTERN) into a code object
        """
        letter = match.group(3)
        args = list(filter(bool, match.group(2).split(";")))
        private = match.group(1)
        if (letter == 'H' or letter == 'f') and not private:
            return CursorPosition(
                row=int(args[0]) if len(args) > 0 else 1,
                column=int(args[1]) if len(args) > 1 else 1
            )
        elif letter == 'm' and not private:
            return GraphicRendition(int(arg) for arg in args)
        else:
            return AnsiCode(letter, args, private)

    @staticmethod
    def split(string):
        """
        Generator that splits a string into a list.
        Each element can either be a simple string or an AnsiCode object
        """
        while string:
            match = AnsiCode.CSI_PATTERN.search(string)
            if not match:
                yield string
                break
            if match.start():
                yield string[:match.start()]
            yield AnsiCode._from_match(match)
            string = string[match.end():]

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __ne__(self, other):
        return not (self == other)


class CursorPosition(AnsiCode):
    """
    Class representing a CUP (Cursor position) code.

    Coordinates are 1-indexed.
    """
    def __init__(self, *args, **kwargs):
        AnsiCode.__init__(self, "H")
        if "x" in kwargs and "y" in kwargs:
            self.column = kwargs["x"]
            self.row = kwargs["y"]
        elif "column" in kwargs and "row" in kwargs:
            self.column = kwargs["column"]
            self.row = kwargs["row"]
        elif len(args) == 2:
            self.column = args[0]
            self.row = args[1]
        else:
            self.column = self.row = 1

    def args(self):
        return [self.row, self.column]


class _GraphicRendition_Meta(type):
    """
    Sets up shortcuts to access standard colors
    """
    @staticmethod
    def __new__(meta, name, bases, dct):
        cls = type.__new__(meta, name, bases, dct)
        for name, value in six.iteritems(vars(cls.Color)):
            if name[0].isupper():
                setattr(cls, name, cls.Color(value))
        return cls


class GraphicRendition(six.with_metaclass(_GraphicRendition_Meta, AnsiCode)):
    """
    Class representing a SGR (Select Graphic Rendition) code.
    """
    Reset       = 0 # Flag to reset all formatting
    Bold        = 1
    Italic      = 3
    Underline   = 4
    Blink       = 5
    Negative    = 7

    class Color(object):
        """
        Class representing an indexed color (8 colors).

        Bright colors result into the non-standard 90 group.
        """
        Black   = 0
        Red     = 1
        Green   = 2
        Yellow  = Red | Green
        Blue    = 4
        Magenta = Blue | Red
        Cyan    = Blue | Green
        White   = Red | Green | Blue
        ResetColor = 9

        def __init__(self, color=ResetColor, background=False, bright=False):
            self.color = color
            self.background = background
            self.bright = bright

        def __repr__(self):
            base = 30
            if self.bright and self.color != GraphicRendition.Color.ResetColor:
                base = 90
            if self.background:
                base += 10
            base += self.color
            return str(base)

    class Color256(object):
        """
        Class representing a non-standard indexed color (256 colors)
        """
        def __init__(self, index, background=False):
            self.color = index
            self.background = background

        def __repr__(self):
            base = 4 if self.background else 3
            return "%s8;5;%s" % (base, self.color)

    class ColorRGB(object):
        """
        Class representing a non-standard 24bit RGB color
        """
        def __init__(self, r, g, b, background=False):
            self.r = r
            self.g = g
            self.b = b
            self.background = background

        def __repr__(self):
            base = 4 if self.background else 3
            return "%s8;2;%s;%s;%s" % (base, self.r, self.g, self.b)

    @staticmethod
    def Background(color):
        """
        Turns a color object into a background color
        """
        bg = copy.copy(color)
        bg.background = True
        return bg

    def __init__(self, *args):
        AnsiCode.__init__(self, "m")

        def _issequence(x):
            try:
                iter(x)
                return True
            except:
                return False

        if len(args) == 1 and _issequence(args[0]):
            flags = list(args[0])
        else:
            flags = list(args)

        def pop_color(index, flags, background, bright):
            if index == 8:
                if bright:
                    return None
                if len(flags) >= 2 and flags[0] == 5:
                    flags.pop(0)
                    return GraphicRendition.Color256(flags.pop(0), background)
                if len(flags) >= 4 and flags[0] == 2:
                    args = flags[1:4] + [background]
                    flags.pop(0)
                    flags.pop(0)
                    flags.pop(0)
                    flags.pop(0)
                    return GraphicRendition.ColorRGB(*args)
                return None
            else:
                return GraphicRendition.Color(index, background, bright)

        self.flags = []

        while flags:
            flag = flags.pop(0)
            if type(flag) is not int:
                self.flags.append(flag)
            elif flag >= 30 and flag < 40:
                color = pop_color(flag % 10, flags, False, False)
                if color:
                    self.flags.append(color)
            elif flag >= 40 and flag < 50:
                color = pop_color(flag % 10, flags, True, False)
                if color:
                    self.flags.append(color)
            elif flag >= 90 and flag < 100:
                color = pop_color(flag % 10, flags, False, True)
                if color:
                    self.flags.append(color)
            elif flag >= 100 and flag < 110:
                color = pop_color(flag % 10, flags, True, True)
                if color:
                    self.flags.append(color)
            else:
                self.flags.append(flag)

    @staticmethod
    def reverse(flag):
        """
        Returns a flag that undoes the given one
        """
        if type(flag) is not int:
            if flag.background:
                return 49
            return 39
        elif flag == 1:
            return 22
        elif flag == 22:
            return 1
        elif flag in six.moves.range(1, 10):
            return 20 + flag
        elif flag in six.moves.range(21, 30):
            return flag - 20
        elif flag in six.moves.range(30, 40) or flag in six.moves.range(90, 100):
            return 39
        elif flag in six.moves.range(40, 50) or flag in six.moves.range(100, 110):
            return 49
        return 0

    def args(self):
        return self.flags


# Shorthand for GraphicRendition
SGR = GraphicRendition


class common(object):
    hide_cursor  = AnsiCode("l", [25], "?")
    show_cursor  = AnsiCode("h", [25], "?")
    alt_screen   = AnsiCode("h", [47], "?")
    main_screen  = AnsiCode("l", [47], "?")
    clear_screen = AnsiCode("J", [2])


class AnsiRenderer(object):
    """
    Class with a simplified interface to render ANSI codes to a file object
    """

    def __init__(self, output=sys.stdout):
        self.output = output

    @contextmanager
    def context_nocursor(self):
        """
        Context manager that disables the text cursor
        """
        self.ansi(common.hide_cursor)
        try:
            yield
        finally:
            self.ansi(common.show_cursor)

    @contextmanager
    def context_alt_screen(self):
        """
        Context manager that switch to an alternate screen
        """
        self.ansi(common.alt_screen)
        try:
            yield
        finally:
            self.ansi(common.main_screen)

    @contextmanager
    def context_sgr(self, *args, **kwargs):
        """
        Context manager that changes the formatting

        with hard_reset clears all formatting on exit
        """
        sgr = SGR(*args)
        hard_reset = kwargs.pop("hard_reset", False)
        self.ansi(sgr)
        try:
            yield
        finally:
            if hard_reset:
                self.ansi(SGR(SGR.Reset))
            else:
                self.ansi(SGR([SGR.reverse(flag) for flag in sgr.flags]))

    @contextmanager
    def rendering_context(self):
        """
        Context manager for a clean rendering environment
        """
        with keyboard_interrupt(), self.context_alt_screen(), self.context_nocursor():
            self.clear()
            self.ansi(SGR(SGR.Reset))
            yield

    def ansi(self, *args, **kwargs):
        """
        Prints an ANSI code to the output
        """
        if not kwargs and len(args) == 1 and isinstance(args[0], AnsiCode):
            code = args[0]
        else:
            code = AnsiCode(*args, **kwargs)
        self.output.write(str(code))
        self.output.flush()

    def move_cursor(self, *args, **kwargs):
        """
        Moves the cursor to the given coordinates
        """
        self.ansi(CursorPosition(*args, **kwargs))

    def clear(self):
        """
        Clears the screen
        """
        self.ansi(common.clear_screen)

    def write(self, *args):
        """
        Writes some text
        """
        self.output.write(" ".join(str(arg) for arg in args))
        self.output.flush()

    def overlay(self, string, start_x=1, start_y=1, tab_width=4):
        """
        Overlays string at the given position

        Space characters are converted into cursor movement, so they will not
        overwrite existing characters, ANSI codes from the strings are preserved
        (They must be in the subset that AnsiCode can parse)
        """
        mover = CharMover(start_x, start_y)
        for item in AnsiCode.split(string):
            if isinstance(item, AnsiCode):
                self.ansi(item)
                #TODO Handle movements
            else:
                for ch in mover.loop(item):
                    if mover.moved:
                        self.move_cursor(mover.x, mover.y)
                    self.output.write(ch)
        self.output.flush()


class CharMover(object):
    """
    Translate space characters into position changes
    """

    def __init__(self, start_x, start_y, tab_width=4):
        self.x = self.start_x = start_x
        self.y = self.start_y = start_y
        self.tab_width = tab_width
        self.moved = False

    def move(self, x, y, moved=True):
        """
        Changes the current position and "moved" status
        """
        self.x = x
        self.y = y
        self.moved = moved or self.moved

    def loop(self, string):
        """
        Generator that parses the string and yields non-space characters.
        Space characters cause x, y to change
        """
        for ch in string:
            if ch == ' ':
                self.x += 1
                self.moved = True
            elif ch == '\n':
                self.x = self.start_x
                self.y += 1
                self.moved = True
            elif ch == '\v':
                self.x += 1
                self.y += 1
                self.moved = True
            elif ch == '\r':
                self.x = self.start_x
                self.moved = True
            elif ch == '\t':
                self.x += self.tab_width
                self.moved = True
            else:
                yield ch
                self.moved = False
                self.x += 1


@contextmanager
def keyboard_interrupt():
    """
    Context manager that exits cleanly on KeyboardInterrupt
    """
    try:
        yield
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python
#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from distutils.core import setup


def get_packages(root):
    if type(root) is str:
        root = [root]
    return sum((
        get_packages(root + [subdir])
        for subdir in
        os.listdir(os.path.join(*root))
        if os.path.isdir(os.path.join(*(root + [subdir])))
    ), [".".join(root)])


setup(
    name="patsi",
    version="0.3",
    description="Python ANSI Terminal Styling Interface",
    long_description="""A Python library to handle ANSI style codes and documents containing colored ASCII art.""",
    author="Mattia Basaglia",
    author_email="mattia.basaglia@gmail.com",
    url="https://github.com/mbasaglia/Python-Ansi-Terminal-Styling-Interface",
    packages=get_packages("patsi"),
    scripts=["patsi-render.py"],
    license="GPLv3+",
    platforms=["any"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Terminals",
        "Operating System :: OS Independent",
    ],
)



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import six
import curses
import curses.ascii
from patsi.widgets.geometry import Point, Rect


class Event(object):
    TYPE_TEXT       = 1
    TYPE_KEY        = 2
    TYPE_RESIZE     = 3
    TYPE_MOUSE      = 4
    TYPE_FOCUS      = 5

    def __init__(self, type, **kwargs):
        self.type = type
        self.propagate = True
        for attr, val in six.iteritems(kwargs):
            setattr(self, attr, val)

    def accept(self):
        self.propagate = False

    @staticmethod
    def text_event(char):
        return Event(Event.TYPE_TEXT, char=char)

    @staticmethod
    def key_event(key):
        return Event(Event.TYPE_KEY, key=key)

    @staticmethod
    def resize_event(bounds):
        return Event(Event.TYPE_RESIZE, bounds=bounds)

    @staticmethod
    def mouse_event(pos, buttons):
        return Event(Event.TYPE_MOUSE, pos=pos, buttons=buttons)

    @staticmethod
    def focus_event(focus):
        return Event(Event.TYPE_FOCUS, focus=focus)


class Widget(object):
    def __init__(self, parent, window=None):
        self.parent = parent
        self.children = []
        self.focus_child = None
        self.active = True
        self.window = window

        if parent is not None:
            bounds = self.window_bounds_hint(parent.window_bounds())
            if window is None:
                self.window = parent.window.subwin(bounds.height, bounds.width, bounds.y, bounds.x)
            else:
                self.adjust_window_size(bounds)
            parent.children.append(self)
        elif window is None:
            self.window = curses.newwin(24, 80)

    def focus(self, *args):
        if not self.active:
            return

        if len(args) == 0 or args[0] is self:
            if self.parent:
                self.parent.focus(self)
            return

        if self.focus_child == args[0]:
            return

        if self.focus_child:
            self.focus_child.event(Event.focus_event(False))

        self.focus_child = args[0]

        if self.focus_child:
            self.focus_child.event(Event.focus_event(True))

    def has_focus(self):
        return self.active and (not self.parent or self.parent.focus_child == self)

    def window_bounds(self):
        return Rect(
            top_left=Point(*self.window.getbegyx(), reversed=True),
            bottom_right=Point(*self.window.getmaxyx(), reversed=True)
        )

    def window_bounds_hint(self, parent_bounds):
        return parent_bounds

    def loop(self):
        while self.active:
            self.refresh()
            self.get_input()

    def deactivate(self):
        self.active = False

    def get_input(self):
        ch = self.window.getch()
        event = None

        if ch == curses.KEY_RESIZE:
            event = Event.resize_event(self.window_bounds())
        elif ch == curses.KEY_MOUSE:
            try:
                id, x, y, z, bstate = curses.getmouse()
                event = Event.mouse_event(Point(x, y), bstate)
            except curses.error:
                pass
        elif curses.ascii.isprint(ch):
            event = Event.text_event(chr(ch))
        else:
            event = Event.key_event(ch)

        if event is not None:
            self.event(event)

    def event(self, event):
        if event.type == Event.TYPE_RESIZE:
            self.resize_event(event)
            if event.propagate:
                bounds = self.window_bounds()
                for child in self.children:
                    old_bounds = child.window_bounds()
                    new_bounds = child.window_bounds_hint(bounds)
                    if old_bounds != new_bounds:
                        child.event(Event.resize_event(new_bounds))
            return

        if not self.active:
            return

        if event.type == Event.TYPE_TEXT:
            self.text_event(event)
            if self.focus_child and event.propagate:
                self.focus_child.event(event)
        elif event.type == Event.TYPE_KEY:
            self.key_event(event)
            if self.focus_child and event.propagate:
                self.focus_child.event(event)
        elif event.type == Event.TYPE_MOUSE:
            self.mouse_event(event)
            if not event.propagate:
                return
            for child in self.children:
                if child.active and child.window_bounds().contains(event.pos):
                    if self.event_change_focus_child(child):
                        child_bounds = child.window_bounds()
                        relative = event.pos - child_bounds.pos
                        child_event = Event.mouse_event(relative, event.buttons)
                        child.event(child_event)
                        if not child_event.propagate:
                            return
        elif event.type == Event.TYPE_FOCUS:
            self.focus_event(event)

    def event_change_focus_child(self, new):
        self.focus_child = new
        return True

    def text_event(self, event):
        pass

    def key_event(self, event):
        pass

    def resize_event(self, event):
        self.adjust_window_size(event.bounds)

    def mouse_event(self, event):
        self.focus()

    def adjust_window_size(self, bounds):
        if self.window_bounds() != bounds:
            self.window.mvwin(bounds.y, bounds.x)
            self.window.resize(bounds.height, bounds.width)

    def refresh(self):
        self.window.clear()
        self.render()
        self.window.refresh()
        for child in self.children:
            if child.active:
                child.refresh()

    def render(self):
        pass

    def focus_event(self, event):
        if event.focus:
            self.refresh()

    def _draw_border(self):
        self.window.border("|", "|", "-", "-", "+", "+", "+", "+")

    def _draw_string(self, pos, text, *args):
        self.window.addstr(pos.y, pos.x, text, *args)



#
# Copyright (C) 2016-2017 Mattia Basaglia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from .. import _misc


_loaders = {}


def register(loader, file_extension):
    _loaders[file_extension] = loader


def loader(file_extension):
    return _loaders[file_extension]


def loader_for_file(file_name):
    return loader(_misc.extension(file_name))


# TODO Use libmagic?
def load(file, hint=None):
    if type(file) is str:
        with open(file, "r") as file_obj:
            return load(file_obj, hint)

    if hint is not None:
        ldr = loader(hint)
    else:
        ldr = loader_for_file(file.name)

    return ldr.load_file(file)


def formats():
    return sorted(_loaders.keys())
